import aiohttp
import asyncio
from threading import Thread

from .Events import Events
from .Errors import *
from .API_Elements import *
from .Voice import *
from .Gateway import *

class Utility:

	def get_self_user(self):
		return User(self.api("/users/@me", "GET"), self)

	def get_self_guilds(self):
		return [Guild(guild,self) for guild in self.api("/users/@me/guilds", "GET")]

	def send_message(self,channel,**kwargs):
		return Message(self.api(f"/channels/{channel}/messages", "POST", json=kwargs),self)

	def get_guild(self,guild_id):
		return Guild(self.api(f"/guilds/{guild_id}","GET"),self)

	def get_channel(self,channel_id):
		return Channel(self.api(f"/channels/{channel_id}","GET"),self)

	def get_user(self,user_id):
		return User(self.api(f"/users/{user_id}"), self)

	def get_invite(self, invite_code):
		return Invite(self.api(f"/invites/{invite_code}","GET", params={"with_counts":"true"}),self)

class Bot(Thread,Utility,Events):

	def __init__(self,token,api_sleep=0.05):
		self.events_list = {}
		Events.__init__(self)
		Thread.__init__(self)
		Utility.__init__(self)
		Bot_Element.__init__(self,{},self)
		self.token=token
		self.api_sleep = api_sleep
		self.api_url="https://discord.com/api"
		self.events = {}
		self.in_wait_voices = []

	def def_event(self,event,name):
		def add_event(function):
			self.events_list[event] = [name,function]
			return
		return add_event

	def event(self,arg):

		def add_event(function):
			self.events[arg]=function
			return

		if type(arg) == str:
			return add_event

		self.events[arg.__name__] = arg
		return

	def get_element(self, element, search):
		try:
			for x in element:
				if str(x.id) == str(search):
					return x
			return
		except:...

	def set_element(self, element, new):
		try:
			for i in range(len(element)):
				if str(element[i].id) == str(new.id):
					element[i] = new
					return
		except:...

	async def api_call(self,path, method="GET", **kwargs):
		defaults = {
			"headers": {
				"Authorization": f"Bot {self.token}",
				"User-Agent": "Bot"
			}
		}
		kwargs = dict(defaults, **kwargs)
		async with aiohttp.ClientSession() as session:
			async with session.request(method, self.api_url+path,**kwargs) as response:
				try:
					assert response.status in [200,204]
					if response.status == 200:
						return await response.json()
				except:
					if response.status == 400:
						return BadRequestError()
					elif response.status == 403:
						return PermissionsError()
					else:
						return Error()

	def api(self, path, method="GET", **kwargs):
		loop = asyncio.new_event_loop()
		output = loop.run_until_complete(self.api_call(path,method,**kwargs))
		loop.run_until_complete(asyncio.sleep(self.api_sleep))
		loop.close()
		if output and isinstance(output,Error):
			raise output
		else:
			return output

	async def begin(self):
		response = await self.api_call("/gateway")
		await self.__main(response["url"])

	async def __main(self,url):
		events = self.events_list
		gateway = Gateway(f"{url}?v=6&encoding=json", self.token)
		self.gateway = gateway
		async for data in gateway.connect():
			if data["op"] == 0:
				if data["t"] in events:
					event = events[data["t"]]
					output = event[1](self,data["d"])
					if event[0] in self.events:
						x = Thread(target=self.events[event[0]],args=(output,))
						x.start()
			if self.in_wait_voices:
				for voice in self.in_wait_voices:
					if voice["guild_id"] not in self.voices:
						x = Voice(voice,self)
						self.voices[voice["guild_id"]] = x
						asyncio.create_task(x.run())
				self.in_wait_voices = []

	def run(self):
		self.loop = asyncio.new_event_loop()
		print("Starting Bot")
		try:
			self.loop.run_until_complete(self.begin())
		except RuntimeError:
			print("Stopping Bot")

	def stop(self):
		self.gateway.stop()


from .API_Elements import *

class OAuth:

	def __init__(self, bot, secret, redirect_uri, scope):
		self.bot = bot
		self.secret = secret
		self.id = bot.get_self_user().id
		self.redirect_uri = redirect_uri
		self.scope = scope

	def get_url(self):
		return f"https://discord.com/api/oauth2/authorize?client_id={self.id}&redirect_uri={self.redirect_uri}&response_type=code&scope={'%20'.join(self.scope.split())}"

	def get_token(self, code):
		data = {
			"client_id": self.id,
			"client_secret": self.secret,
			"grant_type": "authorization_code",
			"code": code,
			"redirect_uri": self.redirect_uri,
			"scope": self.scope
		}

		return self.__bot.api("/oauth2/token","POST",data=data)

	def __request_token(self,token,url):
		headers = {
			"Authorization": f"Bearer {json.loads(token)['access_token']}"
		}

		return self.__bot.api(url,"GET",headers=headers)

	def get_user(self,token):
		if "identify" in self.scope:
			return User(self.__request_token(token,"/users/@me"),self.bot)
		return "Invalid Scope"

	def get_guilds(self,token):
		if "guilds" in self.scope:
			return [Guild(guild,self.bot) for guild in self.__request_token(token,"/users/@me/guilds")]
		return "Invalid Scope"

	def add_guild_member(self, token, guild_id, user_id):
		if "guilds.join" in self.scope:
			return Member({**self.bot.api_call(f"/guilds/{guild_id}/members/{user_id}","PUT",json=json.loads(token)["access_token"]),"guild_id":guild_id})
		return "Invalid Scope"


from .API_Elements import *

class Events:
	def __init__(self):

		@self.def_event("GUILD_CREATE","guild_create")
		class Event(Guild):

			def __init__(self, bot, data):
				Guild.__init__(self, data, bot)

				guild = bot.get_element(bot.guilds,self.id)
				if guild:
					for x,y in self.__dict__.items():
						setattr(guild,x,y)
				else:
					bot.guilds.append(self)

		@self.def_event("MESSAGE_CREATE","on_message")
		class Event(Message):

			def __init__(self, bot, data):
				Message.__init__(self, data, bot)

		@self.def_event("MESSAGE_UPDATE","message_update")
		class Event(Message):

			def __init__(self, bot, data):
				Message.__init__(self, data, bot)

		@self.def_event("MESSAGE_DELETE","message_delete")
		class Event(Message):

			def __init__(self, bot, data):
				Message.__init__(self, data, bot)

		@self.def_event("READY","on_ready")
		class Event(Bot_Element):

			def __init__(self, bot, data):
				Bot_Element.__init__(self,data,bot)
				self.version = data["v"]

				for x,y in self.__dict__.items():
					setattr(bot,x,y)

		@self.def_event("MESSAGE_REACTION_ADD","reaction_add")
		class Event(Member):

			def __init__(self,bot,data):
				if "member" in data:
					Member.__init__(self,{**data["member"],"guild_id":data["guild_id"]},bot)
				self.emoji = Emoji(data["emoji"])
				self.user_id = data["user_id"]
				self.channel_id = data["channel_id"]
				self.message_id = data["message_id"]
				self.__bot = bot

			@property
			@Cache
			def message(self):
				return Message(self.__bot.api(f"/channels/{self.channel_id}/messages/{self.message_id}"),self.__bot)

			def delete(self):
				if self.id == self.__bot.user.id:
					self.message.delete_self_reaction(self.emoji.name)
				else:
					self.message.delete_reaction(self.emoji.name,user_id=self.id)

		@self.def_event("MESSAGE_REACTION_REMOVE","reaction_remove")
		class Event:

			def __init__(self, bot, data):
				self.user_id = data["user_id"]
				self.channel_id = data["channel_id"]
				self.message_id = data["message_id"]
				self.guild_id = data.get("guild_id")
				self.emoji = Emoji(data["emoji"])
				self.__bot = bot

			@property
			@Cache
			def message(self):
				return Message(self.__bot.api(f"/channels/{self.channel_id}/messages/{self.message_id}"),self.__bot)

		@self.def_event("CHANNEL_CREATE","channel_create")
		class Event(Channel):

			def __init__(self, bot, data):
				Channel.__init__(self, data, bot)

				if self.type == 1:
					if bot.get_element(bot.private_channels, self.id):
						bot.set_element(bot.private_channels, self)
					else:
						bot.private_channels.append(self)
				else:
					self.guild = bot.get_element(bot.guilds, self.guild_id)
					self.guild.channels.append(self)

		@self.def_event("CHANNEL_UPDATE","channel_update")
		class Event(Channel):

			def __init__(self, bot, data):
				Channel.__init__(self, data, bot)

				self.guild = bot.get_element(bot.guilds, self.guild_id)
				bot.set_element(self.guild.channels, self)

		@self.def_event("CHANNEL_DELETE","channel_delete")
		class Event(Channel):

			def __init__(self, bot, data):
				Channel.__init__(self, data, bot)

				self.guild = bot.get_element(bot.guilds, self.guild_id)
				self.guild.channels.remove(bot.get_element(self.guild.channels, self.id))

		@self.def_event("GUILD_MEMBER_ADD","member_join")
		class Event(Member):

			def __init__(self, bot, data):
				Member.__init__(self,data,bot)

				self.guild = bot.get_element(bot.guilds, self.guild_id)
				self.guild.members.append(self)

		@self.def_event("GUILD_MEMBER_UPDATE","member_update")
		class Event(Member):
			def __init__(self, bot, data):
				Member.__init__(self,data,bot)

				self.guild = bot.get_element(bot.guilds, self.guild_id)
				bot.set_element(self.guild.members,self)

		@self.def_event("GUILD_MEMBER_REMOVE","member_quit")
		class Event(User):

			def __init__(self, bot, data):
				User.__init__(self,data["user"],bot)
				self.guild_id = data["guild_id"]

				self.guild = bot.get_element(bot.guilds, self.guild_id)
				self.guild.members.remove(bot.get_element(self.guild.members, self.id))

		@self.def_event("GUILD_ROLE_CREATE","role_create")
		class Event(Role):

			def __init__(self, bot, data):
				Role.__init__(self, data["role"], bot)
				self.guild_id = data["guild_id"]

				self.guild = bot.get_element(bot.guilds, self.guild_id)
				self.guild.roles.append(self)

		@self.def_event("GUILD_ROLE_UPDATE","role_update")
		class Event(Role):
			def __init__(self, bot, data):
				Role.__init__(self, data["role"], bot)
				self.guild_id = data["guild_id"]

				self.guild = bot.get_element(bot.guilds, self.guild_id)
				bot.set_element(self.guild.roles,self)

		@self.def_event("GUILD_ROLE_DELETE","role_delete")
		class Event:

			def __init__(self, bot, data):
				self.guild_id = data["guild_id"]
				self.id = data["role_id"]

				self.guild = bot.get_element(bot.guilds, self.guild_id)
				role = bot.get_element(self.guild.roles, self.id)
				for x,y in role.__dict__.items():
					setattr(self,x,y)
				self.guild.roles.remove(role)

		@self.def_event("INVITE_CREATE","invite_create")
		class Event(Invite):

			def __init__(self, bot, data):
				Invite.__init__(self, data, bot)
				self.channel_id = data["channel_id"]
				self.guild_id = data.get("guild_id",None)

				self.guild = bot.get_element(bot.guilds, self.guild_id)
				self.channel = bot.get_element(self.guild.channels, self.channel_id)
				self.channel.invites.append(self)

		@self.def_event("INVITE_DELETE","invite_delete")
		class Event(Invite):

			def __init__(self, bot, data):
				Invite.__init__(self, data, bot)
				self.channel_id = data["channel_id"]
				self.guild_id = data.get("guild_id",None)

				self.guild = bot.get_element(bot.guilds, self.guild_id)
				self.channel = bot.get_element(self.guild.channels, self.channel_id)
				for invite in channel.invites:
					if invite.code == self.code:
						for x,y in invite.__dict__.items():
							setattr(self,x,y)
						self.channel.invites.remove(invite)
						break

'''
		@self.def_event("VOICE_STATE_UPDATE","")
		class Event:

			def __init__(self, bot, data):
				self.voice = bot.voices.get(data["guild_id"])
				self.member = Member(data["member"],bot)
				self.session_id = data.get("session_id")
				self.__bot = bot

			async def run(self):
				if self.member.id == self.__bot.user.id:
					if self.voice:
						if not self.voice.session_id:
							self.voice.session_id = self.session_id
							self.voice.state += 1
							if self.voice.state == 2:
								await self.voice.run()

		@self.def_event("VOICE_SERVER_UPDATE","")
		class Event:

			def __init__(self, bot, data):
				self.voice = bot.voices[data["guild_id"]]
				self.token = data["token"]
				self.endpoint = data["endpoint"]

			async def run(self):
				if not self.voice.token:
					self.voice.token = self.token
					self.voice.endpoint = self.endpoint.replace(":80","")
					self.voice.state += 1
					if self.voice.state == 2:
						await self.voice.run()
'''

		


class Cache:
	def __init__(self, func):
		self.func = func
		self.result = None

	def __call__(self,ref):
		if self.result:
			return self.result
		result = self.func(ref)
		self.result = result
		return result

class API_Element:

	def to_json(self):
		output = {}
		for x,y in self.__dict__.items():
			if x.endswith("__bot"):continue
			if y != None:
				if type(y) == list:
					e=[]
					for p in y:
						if isinstance(p,API_Element):
							p = p.to_json()
						e += [p]
					y = e
				if isinstance(y,API_Element):
					y = y.to_json()
				output[x]=y
		return output


class Bot_Element:

	def __init__(self, bot_element, bot):
		self.user = User(bot_element.get("user",{}), bot)
		self.guilds = [Guild(guild, bot) for guild in bot_element.get("guilds",[])]
		self.relationships = bot_element.get("relationships",[])
		self.private_channels = bot_element.get("private_channels",[])
		self.presences = bot_element.get("presences",[])
		self.voices = {}

class Guild:

	def __init__(self, guild, bot):
		self.id = guild.get("id")
		self.name = guild.get("name")
		self.icon = guild.get("icon")
		self.splash = guild.get("splash")
		self.discovery_splash = guild.get("discovery_splash")
		self.owner = guild.get("owner")
		self.owner_id = guild.get("owner_id")
		self.permissions = guild.get("permissions")
		self.region = guild.get("region")
		self.afk_channel_id = guild.get("afk_channel_id")
		self.afk_timeout = guild.get("afk_timeout")
		self.embed_enabled = guild.get("embed_enabled")
		self.embed_channel_id = guild.get("embed_channel_id")
		self.verification_level = guild.get("verification_level")
		self.default_message_notifications = guild.get("default_message_notifications")
		self.explicit_content_filter = guild.get("explicit_content_filter")
		self.roles = [Role(role,bot) for role in guild.get("roles",[])]
		self.emojis = [Emoji(emoji) for emoji in guild.get("emojis",[])]
		self.features = guild.get("features",[]) #To Do
		self.mfa_level = guild.get("mfa_level")
		self.application_id = guild.get("application_id")
		self.widget_enabled = guild.get("widget_enabled")
		self.widget_channel_id = guild.get("widget_channel_id")
		self.system_channel_id = guild.get("system_channel_id")
		self.system_channel_flags = guild.get("system_channel_flags")
		self.rules_channel_id = guild.get("rules_channel_id")
		self.joined_at = guild.get("joined_at")
		self.large = guild.get("large")
		self.unavailable = guild.get("unavailable")
		self.member_count = guild.get("member_count")
		self.voice_states = guild.get("voice_states")
		self.members = [Member(member,bot) for member in guild.get("members",[])]
		self.channels = [Channel(channel,bot) for channel in guild.get("channels",[])]
		self.presences = guild.get("presences",[]) # To Do
		self.max_presences = guild.get("max_presences")
		self.max_members = guild.get("max_members")
		self.max_video_channel_users = guild.get("max_video_channel_users")
		self.vanity_url_code = guild.get("vanity_url_code")
		self.description = guild.get("description")
		self.banner = guild.get("banner")
		self.premium_tier = guild.get("premium_tier")
		self.premium_subscription_count = guild.get("premium_subscription_count")
		self.preferred_locale = guild.get("preferred_locale")
		self.public_updates_channel_id = guild.get("public_updates_channel_id")
		self.approximate_member_count = guild.get("approximate_member_count")
		self.approximate_presence_count = guild.get("approximate_presence_count")
		self.__bot = bot

	def __repr__(self):
		if self.name:
			return self.name
		elif self.id:
			return self.id
		else:
			return "Guild"

	def get_channels(self):
		channels = self.__bot.api(f"/guilds/{self.id}/channels")
		return [Channel(channel,self.__bot) for channel in channels]

	def get_roles(self):
		roles = self.__bot.api(f"/guilds/{self.id}/roles")
		return [Role(role,self.__bot) for role in roles]

	def get_invites(self):
		invites = self.__bot.api(f"/guilds/{self.id}/invites")
		return [Invite(invite,self.__bot) for invite in invites]

	def get_members(self, limit=100, after=0):
		members = self.__bot.api(f"/guilds/{self.id}/members","GET",params={"limit":limit,"after":after})
		return [Member({**member,"guild_id":self.id},self.__bot) for member in members]

	def get_member(self,user_id):
		return Member({**self.__bot.api(f"/guilds/{self.id}/members/{user_id}"),"guild_id":self.id},self.__bot)

	def get_bans(self):
		bans = self.__bot.api(f"/guilds/{self.id}/bans")
		return [Ban(ban,self.__bot) for ban in bans]

	def get_ban(self, user_id):
		return Ban(self.__bot.api(f"/guilds/{self.id}/bans/{user_id}"),self.__bot)

	def create_channel(self,**kwargs):
		''' kwargs : https://discord.com/developers/docs/resources/guild#create-guild-channel '''
		return Channel(self.__bot.api(f"/guilds/{self.id}/channels", "POST", json=kwargs),self.__bot)

	def create_role(self,**kwargs):
		''' kwargs : https://discord.com/developers/docs/resources/guild#create-guild-role '''
		return Role(self.__bot.api(f"/guilds/{self.id}/roles", "POST", json=kwargs),self.__bot)

class Channel:

	def __init__(self, channel, bot):
		self.id = channel.get("id")
		self.type = channel.get("type")
		self.guild_id = channel.get("guild_id")
		self.position = channel.get("position")
		self.permission_overwrites = [Overwrite(overwrite,bot,self.id) for overwrite in channel.get("permission_overwrites",[])]
		self.name = channel.get("name")
		self.topic = channel.get("topic")
		self.nsfw = channel.get("nsfw")
		self.last_message_id = channel.get("last_message_id")
		self.bitrate = channel.get("bitrate")
		self.user_limit = channel.get("user_limit")
		self.rate_limit_per_user = channel.get("rate_limit_per_user")
		self.recipients = [User(user,bot) for user in channel.get("recipients",[])]
		self.icon = channel.get("icon")
		self.owner_id = channel.get("owner_id")
		self.application_id = channel.get("application_id")
		self.parent_id = channel.get("parent_id")
		self.last_pin_timestamp = channel.get("last_pin_timestamp")
		self.invites = channel.get("invites",[])
		self.__bot = bot

	def __repr__(self):
		if self.name: return self.name
		else: return self.id

	def edit(self,**modifs):
		self.__bot.api(f"/channels/{self.id}","PATCH",json=modifs)

	def send(self,content=None,**kwargs):
		return Message(self.__bot.api(f"/channels/{self.id}/messages", "POST", json={"content":content,**kwargs}),self.__bot)

	def get_messages(self):
		messages = self.__bot.api(f"/channels/{self.id}/messages")
		return [Message(message,self.__bot) for message in messages]

	def get_invites(self):
		invites = self.__bot.api(f"/channels/{self.id}/invites")
		return [Invite(invite,self.__bot) for invite in invites]

	def create_invite(self,**kwargs):
		return Invite(self.__bot.api(f"/channels/{self.id}/invites","POST",json=kwargs),self.__bot)

	def typing(self):
		self.__bot.api(f"/channels/{self.id}/typing","POST")

class Message:

	def __init__(self, message, bot):
		self.id = message.get("id")
		self.channel_id = message.get("channel_id")
		self.guild_id = message.get("guild_id")
		if "author" in message:
			self.author = User(message["author"],bot)
		if "member" in message:
			self.author = Member({**message["member"],"user":{**message["author"]},"guild_id":self.guild_id}, bot)
		self.content = message.get("content")
		self.timestamp = message.get("timestamp")
		self.edited_timestamp = message.get("edited_timestamp")
		self.tts = message.get("tts")
		self.mention_everyone = message.get("mention_everyone")
		self.mentions = [User(mention,bot) for mention in message.get("mentions")]
		self.mentions_roles = message.get("mention_roles")
		self.mention_channels = [Channel(channel,bot) for channel in message.get("mention_channels",[])]
		self.attachments = [Attachment(attachment) for attachment in message.get("attachments",[])]
		self.embeds = [Embed(embed) for embed in message.get("embeds",[])]
		self.reactions = [Reaction(reaction,self.id) for reaction in message.get("reactions",[])]
		self.nonce = message.get("nonce")
		self.pinned = message.get("pinned")
		self.webhook_id = message.get("webhook_id")
		self.type = message.get("type")
		self.activity = message.get("activity") # Object
		self.application = message.get("application") #Object
		self.message_reference = message.get("message_reference") #Object
		self.flags = message.get("flags")
		self.__bot = bot

		self.guild = bot.get_element(bot.guilds, self.guild_id)
		if self.guild:
			self.channel = bot.get_element(self.guild.channels, self.channel_id)
		else:
			self.channel = bot.get_element(bot.private_channels, self.channel_id)

	def __repr__(self):
		return self.content

	def delete(self):
		self.__bot.api(f"/channels/{self.channel_id}/messages/{self.id}","DELETE")

	def edit(self,**modifs):
		self.__bot.api(f"/channels/{self.channel_id}/messages/{self.id}","PATCH",json=modifs)

	def add_reaction(self, reaction):
		self.__bot.api(f"/channels/{self.channel_id}/messages/{self.id}/reactions/{reaction}/@me","PUT")

	def delete_reactions(self):
		self.__bot.api(f"/channels/{self.channel_id}/messages/{self.id}/reactions","DELETE")

	def delete_self_reaction(self, reaction):
		self.__bot.api(f"/channels/{self.channel_id}/messages/{self.id}/reactions/{reaction}/@me","DELETE")

	def delete_reaction(self,reaction,user_id=None):
		if user_id:
			self.__bot.api(f"/channels/{self.channel_id}/messages/{self.id}/reactions/{reaction}/{user_id}","DELETE")
		else:
			self.__bot.api(f"/channels/{self.channel_id}/messages/{self.id}/reactions/{reaction}","DELETE")


class User:

	def __init__(self, user, bot):
		self.bot = user.get("bot")
		self.system = user.get("system")
		self.mfa_enabled = user.get("mfa_enabled")
		self.locale = user.get("locale")
		self.verified = user.get("verified")
		self.email = user.get("email")
		self.flags = user.get("flags")
		self.premium_type = user.get("premium_type")
		self.public_flags = user.get("public_flags")
		self.id = user.get("id")
		self.name = user.get("username")
		self.discriminator = user.get("discriminator")
		self.avatar = user.get("avatar")
		if self.avatar:
			if self.avatar.startswith("a_"):
				avatar_type = ".gif"
			else:
				avatar_type = ".png"
			self.avatar = f"https://cdn.discordapp.com/avatars/{self.id}/{self.avatar}{avatar_type}"
		self.mention = f"<@{self.id}>"
		self.__bot = bot

	def __repr__(self):
		return self.name

	@property
	@Cache
	def dm(self):
		return Channel(self.__bot.api(f"/users/@me/channels","POST",json={"recipient_id":self.id}),self.__bot)

class Member(User):

	def __init__(self, member, bot):
		if "user" in member:
			User.__init__(self,member["user"],bot)
		self.guild_id = member.get("guild_id")
		self.premium_since = member.get("premium_since")
		self.roles = [role for role in member.get("roles",[])]
		self.hoisted_role = member.get("hoisted_role")
		self.mute = member.get("mute")
		self.deaf = member.get("deaf")
		self.nick = member.get("nick")
		self.joined_at = member.get("joined_at")
		self.__bot = bot

	def edit(self, **modifs):
		if hasattr(self,"id"):
			user_id=self.id
			self.__bot.api(f"/channels/{self.guild_id}/messages/{user_id}","PATCH",json=modifs)

	def kick(self):
		if hasattr(self,"id"):
			user_id=self.id
			delete_member = self.__bot.api(f"/guilds/{self.guild_id}/members/{user_id}","DELETE")

	def ban(self, reason=None):
		if hasattr(self,"id"):
			user_id=self.id
			self.__bot.api(f"/guilds/{self.guild_id}/bans/{user_id}","PUT", json={"reason":reason})

	def add_role(self, role):
		if hasattr(self,"id"):
			user_id=self.id
			self.__bot.api(f"/guilds/{self.guild_id}/members/{user_id}/roles/{role.id}","PUT")

	def remove_role(self, role):
		if hasattr(self,"id"):
			user_id=self.id
			self.__bot.api(f"/guilds/{self.guild_id}/members/{user_id}/roles/{role.id}","DELETE")

class Reaction:

	def __init__(self,reaction,message_id):
		self.count = reaction.get("count")
		self.me = reaction.get("me")
		self.emoji = Emoji(reaction["emoji"])
		self.message_id = message_id

class Emoji:

	def __init__(self,emoji):
		self.name = emoji.get("name")
		self.id = emoji.get("id")
		self.roles = emoji.get("roles")
		self.user = emoji.get("user")
		self.require_colons = emoji.get("require_colons")
		self.managed = emoji.get("managed")
		self.animated = emoji.get("animated")
		self.available = emoji.get("available")


class Role:

	def __init__(self, role, bot):
		self.id = role.get("id")
		self.name = role.get("name")
		self.color = role.get("color")
		self.hoist = role.get("hoist")
		self.position = role.get("position")
		self.permissions = role.get("permissions")
		self.managed = role.get("managed")
		self.mentionable = role.get("mentionable")
		self.guild_id = role.get("guild_id")
		self.__bot = bot

	def __repr__(self):
		return self.name

	def delete(self):
		self.__bot.api(f"/channels/{self.guild_id}/roles/{self.id}","DELETE")

	def edit(self,**modifs):
		self.__bot.api(f"/channels/{self.guild_id}/messages/{self.id}","PATCH",json=modifs)


class Attachment:

	def __init__(self,attachment):
		self.id = attachment.get("id")
		self.filename = attachment.get("filename")
		self.size = attachment.get("size")
		self.url = attachment.get("url")
		self.proxy_url = attachment.get("proxy_url")
		self.height = attachment.get("height")
		self.width = attachment.get("width")

class Allowed_Mentions:

	def __init__(self,mentions):
		self.parse = mentions.get("parse")
		self.roles = mentions.get("roles")
		self.users = mentions.get("users")

class Embed(API_Element):

	def __init__(self,embed):
		self.title = embed.get("title")
		self.type = embed.get("type")
		self.description = embed.get("description")
		self.url = embed.get("url")
		self.timestamp = embed.get("timestamp")
		self.color = embed.get("color")
		self.footer = Embed_Footer(embed.get("footer",{}))
		self.image = Embed_Image(embed.get("image",{}))
		self.thumbnail = Embed_Image(embed.get("thumbnail",{}))
		self.video = Embed_Image(embed.get("video",{}))
		self.provider = Embed_Provider(embed.get("provider",{}))
		self.author = Embed_Author(embed.get("author",{}))
		self.fields = [Embed_Field(field) for field in embed.get("fields",[])]

	def add_field(self,**kwargs):
		self.fields.append(Embed_Field(kwargs))
		return self.fields[-1]

class Embed_Image(API_Element):

	def __init__(self,image):
		self.url = image.get("url")
		self.proxy_url = image.get("proxy_url")
		self.height = image.get("height")
		self.width = image.get("width")

class Embed_Field(API_Element):

	def __init__(self,field):
		self.name = field.get("name")
		self.value = field.get("value")
		self.inline = field.get("inline")

class Embed_Footer(API_Element):

	def __init__(self,footer):
		self.text = footer.get("text")
		self.icon_url = footer.get("icon_url")
		self.proxy_icon_url = footer.get("proxy_icon_url")

class Embed_Provider(API_Element):

	def __init__(self,provider):
		self.name = provider.get("name")
		self.url = provider.get("url")

class Embed_Author(API_Element):

	def __init__(self,author):
		self.name = author.get("name")
		self.url = author.get("url")
		self.icon_url = author.get("icon_url")
		self.proxy_icon_url = author.get("proxy_icon_url")

class Invite:

	def __init__(self, invite, bot):
		self.code = invite.get("code")
		self.url = f"https://discord.gg/{self.code}"
		self.guild = None
		if "guild" in invite:
			self.guild = Guild(invite["guild"], bot)
		self.channel = None
		if "channel" in invite:
			self.channel = Channel(invite["channel"], bot)
		self.inviter = None
		if "inviter" in invite:
			self.inviter = User(invite["inviter"], bot)
		self.target_user = None
		if "target_user" in invite:
			self.channel = User(invite["target_user"], bot)
		self.target_user_type = invite.get("target_user_type")
		self.approximate_presence_count = invite.get("approximate_presence_count")
		self.approximate_member_count = invite.get("approximate_member_count")
		self.max_age = invite.get("max_age")
		self.max_uses = invite.get("max_uses")
		self.temporary = invite.get("temporary")
		self.uses = invite.get("uses")
		self.created_at = invite.get("created_at")
		self.__bot = bot

	def __repr__(self):
		return self.url

	def delete(self):
		self.__bot.api(f"/invites/{self.code}","DELETE")

class Ban:

	def __init__(self,ban,bot):
		self.reason = ban.get("reason")
		self.user = User(ban["user"])
		self.__bot = bot

	def pardon(self, guild_id):
		self.__bot.api(f"/guilds/{guild_id}/bans/{self.user.id}","DELETE")

class Overwrite:

	def __init__(self,overwrite,bot,channel_id):
		self.id = overwrite.get("id")
		self.type = overwrite.get("type")
		self.allow = overwrite.get("allow")
		self.deny = overwrite.get("deny")
		self.channel_id = channel_id
		self.__bot = bot

	def edit(self,**modifs):
		self.__bot.api(f"/channels/{self.channel_id}/permissions/{self.id}","PUT",json=modifs)

	def delete(self):
		self.__bot.api(f"/channels/{self.channel_id}/permissions/{self.id}","DELETE")



class Perm:

	def __init__(self,number):
		self.number = 2**number

	def __eq__(self,n):
		if n & self.number == self.number:
			return True
		else:
			return False

create_instant_invite = Perm(0)
kick_members = Perm(1)
ban_members = Perm(2)
administrator = Perm(3)
manage_channels = Perm(4)
manage_guild = Perm(5)
add_reactions = Perm(6)
view_audit_log = Perm(7)
priority_speaker = Perm(8)
stream = Perm(9)
view_channel = Perm(10)
send_messages = Perm(11)
send_tts_messages = Perm(12)
manage_messages = Perm(13)
embed_links = Perm(14)
attach_files = Perm(15)
read_message_history = Perm(16)
mention_everyone = Perm(17)
use_external_emojis = Perm(18)
view_guild_insights = Perm(19)
connect = Perm(20)
speak = Perm(21)
mute_members = Perm(22)
deafen_members = Perm(23)
move_members = Perm(24)
use_vad = Perm(25)
change_nickname = Perm(26)
manage_nicknames = Perm(27)
manage_roles = Perm(28)
manage_webhooks = Perm(29)
manage_emojis = Perm(30)


import asyncio

from .Gateway import *

class Voice:

	def __init__(self, voice, bot):
		self.guild_id = voice["guild_id"]
		self.channel_id = voice["channel_id"]
		self.mute = voice.get("mute",None)
		self.deaf = voice.get("deaf",None)
		self.state = 0
		self.gateway = None
		#self.client = Voice_Client(self.guild_id,self.channel_id)
		self.__bot = bot

	async def run(self):
		self.loop = asyncio.get_event_loop()
		response = await self.__bot.api_call("/gateway")
		gateway = Gateway(f"{response['url']}?v=6&encoding=json", self.__bot.token)
		self.gateway = gateway
		async for data in gateway.connect():
			if data["op"] == 0:
				if data["t"] == "READY":
					await gateway.send({
						"op":4,
						"d":{
							"guild_id": self.guild_id,
							"channel_id": self.channel_id,
							"mute": self.mute,
							"deaf": self.deaf
					}})
				'''
				if data["t"] == "VOICE_SERVER_UPDATE":
					if data["d"]["guild_id"] == self.guild_id:
						self.client.token = data["d"]["token"]
						self.client.endpoint = data["d"]["endpoint"]
				elif data["t"] == "VOICE_STATE_UPDATE":
					if data["d"]["user_id"] == self.__bot.user.id and data["d"]["guild_id"] == self.guild_id:
						self.client.session_id = data["d"]["session_id"]
				if self.client.session_id and self.client.token and not self.state:
					self.state = 1
					asyncio.create_task(self.client.run())
				'''

	def stop(self):
		#if self.client.gateway:
		#	self.client.stop()
		if self.gateway:
			self.gateway.stop()

'''
class Voice_Client:

	def __init__(self, guild_id, user_id):
		self.guild_id = guild_id
		self.user_id = user_id
		self.session_id = None
		self.token = None
		self.endpoint = None
		self.gateway = None
	
	async def run(self):
		print("aaa")
		gateway = Gateway(f"wss://{self.endpoint[:-3]}/?v=4", "", 8, False)
		self.gateway = gateway
		payload = {
			"op": 0,
			"d": {
				"server_id": self.guild_id,
				"user_id": self.user_id,
				"session_id": self.session_id,
				"token": self.token
		}}
		async for data in gateway.connect(payload):
			if data["op"] == 2:
				print(data)
				self.ip = data["d"]["ip"]
				self.port = data["d"]["port"]
				payload = {
					"op": 1,
					"d": {
						"protocol": "udp",
						"data": {
							"address": self.ip,
							"port": self.port,
							"mode": "xsalsa20_poly1305_lite"
				}}}
				await gateway.send(payload)
			if data["op"] == 4:
				print(data)

	def stop(self):
		self.gateway.stop()
'''


class Error(Exception):

	def __str__(self):
		return f"Unknown Error"

class PermissionsError(Error):

	def __str__(self):
		return f"You do not have permission to do this action"

class BadRequestError(Error):

	def __str__(self):
		return f"Your query was incomplete or bad"


import websockets
import json
import asyncio

class Gateway:

	def __init__(self, url, token, auth_op = 10, events_code = 0):
		self.url = url
		self.token = token
		self.auth_op = auth_op
		self.last_sequence = 0
		self.events_code = events_code
		self.session_id = None

	async def connect(self, action = None):
		self.loop = asyncio.get_event_loop()
		ws = await websockets.connect(self.url)
		self.ws = ws
		while True:
			try:
				msg = await self.ws.recv()
			except websockets.exceptions.ConnectionClosedOK as e:
				if e.code == 1001:
					if self.session_id and self.last_sequence:
						payload = {
							"op": 6,
							"d": {
								"token": self.token,
								"session_id": self.session_id,
								"seq": self.last_sequence
						}}
						await self.ws.close()
						self.ws = await websockets.connect(self.url)
						await self.send(payload)
						data = await self.ws.recv()
						self.interval = data["d"]["heartbeat_interval"]
						continue
				return
			data = json.loads(msg)
			x = data.get("s")
			if x:
				self.last_sequence = x
			if data["op"] == self.auth_op:
				self.interval = data["d"]["heartbeat_interval"]
				self.heartbeat = asyncio.create_task(self.__heartbeat(self.ws))
				payload = {
					"op": 2,
					"d": {
						"token": self.token,
						"properties": {
							"$browser": "Piscord",
							"$device": "Piscord"
				}}}
				await self.send(payload)
			if data["op"] == self.events_code:
				if data["t"] == "READY":
					self.session_id = data["d"]["session_id"]
			yield data

	async def __heartbeat(self, ws):
		while True:
			await asyncio.sleep(self.interval / 1000)
			await self.send({"op": 1,"d": self.last_sequence})

	async def send(self, payload):
		await self.ws.send(json.dumps(payload))

	async def _stop(self):
		await self.ws.close()

	def stop(self):
		asyncio.run_coroutine_threadsafe(self._stop(), self.loop)


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Piscord",
    version="1.3.0-1",
    author="Astremy",
    description="Piscord is a python framework to communicate with the Discord api.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Astremy/Piscord",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)